/*
 * Common clock and power management functionality.
 *
 * Copyright (C) 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "cpm-common.h"
#include "cpm-generic.h"



Cpm_chip::Cpm_chip(l4_addr_t addr, Clock_base *clocks[])
: _clocks(clocks), _cpm_regs(addr, clocks)
{
}

const char *
Cpm_chip::clock_type(enum Clock_identifiers clock)
{
  return _clocks[clock]->clock_type();
}

int
Cpm_chip::have_clock(enum Clock_identifiers clock)
{
  return _clocks[clock]->have_clock(_cpm_regs);
}

void
Cpm_chip::start_clock(enum Clock_identifiers clock)
{
  _clocks[clock]->start_clock(_cpm_regs);
}

void
Cpm_chip::stop_clock(enum Clock_identifiers clock)
{
  _clocks[clock]->stop_clock(_cpm_regs);
}

int
Cpm_chip::get_parameters(enum Clock_identifiers clock, uint32_t parameters[])
{
  Clock_divided_base *clk = dynamic_cast<Clock_divided_base *>(_clocks[clock]);

  if (clk != NULL)
    return clk->get_parameters(_cpm_regs, parameters);
  else
    return 0;
}

int
Cpm_chip::set_parameters(enum Clock_identifiers clock, int num_parameters, uint32_t parameters[])
{
  Clock_divided_base *clk = dynamic_cast<Clock_divided_base *>(_clocks[clock]);

  if (clk != NULL)
    return clk->set_parameters(_cpm_regs, num_parameters, parameters);
  else
    return 0;
}

uint8_t
Cpm_chip::get_source(enum Clock_identifiers clock)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(_clocks[clock]);

  if (clk != NULL)
    return clk->get_source(_cpm_regs);
  else
    return 0;
}

enum Clock_identifiers
Cpm_chip::get_source_clock(enum Clock_identifiers clock)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(_clocks[clock]);

  if (clk != NULL)
    return clk->get_source_clock(_cpm_regs);
  else
    return Clock_none;
}

void
Cpm_chip::set_source_clock(enum Clock_identifiers clock, enum Clock_identifiers source)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(_clocks[clock]);

  if (clk != NULL)
    clk->set_source_clock(_cpm_regs, source);
}

void
Cpm_chip::set_source(enum Clock_identifiers clock, uint8_t source)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(_clocks[clock]);

  if (clk != NULL)
    clk->set_source(_cpm_regs, source);
}

uint64_t
Cpm_chip::get_source_frequency(enum Clock_identifiers clock)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(_clocks[clock]);

  if (clk != NULL)
    return clk->get_source_frequency(_cpm_regs);
  else
    return 0;
}

uint64_t
Cpm_chip::get_frequency(enum Clock_identifiers clock)
{
  return _clocks[clock]->get_frequency(_cpm_regs);
}

int
Cpm_chip::set_frequency(enum Clock_identifiers clock, uint64_t frequency)
{
  Clock_divided_base *clk = dynamic_cast<Clock_divided_base *>(_clocks[clock]);

  if (clk != NULL)
    return clk->set_frequency(_cpm_regs, frequency);
  else
    return 0;
}
