/*
 * DMA support for the X1600.
 *
 * Copyright (C) 2021, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/cpm-x1600.h>
#include <l4/devices/dma-x1600.h>
#include <l4/devices/hw_mmio_register_block.h>

#include <l4/sys/cache.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/util/util.h>

#include <systypes/thread.h>

#include <stdio.h>



enum Global_regs
{
  Dma_control                 = 0x1000, // DMAC
  Dma_irq_pending             = 0x1004, // DIRQP
  Dma_doorbell                = 0x1008, // DDB
  Dma_doorbell_set            = 0x100c, // DDS
};

enum Channel_regs
{
  Dma_source                  = 0x00, // DSA
  Dma_destination             = 0x04, // DTA
  Dma_transfer_count          = 0x08, // DTC
  Dma_request_source          = 0x0c, // DRT
  Dma_control_status          = 0x10, // DCS
  Dma_command                 = 0x14, // DCM
  Dma_descriptor_address      = 0x18, // DDA
  Dma_stride                  = 0x1c, // DSD
};

enum Dma_control_bits : unsigned
{
  Dma_fast_msc_transfer       = 0x80000000, // FMSC
  Dma_fast_ssi_transfer       = 0x40000000, // FSSI
  Dma_fast_tssi_transfer      = 0x20000000, // FTSSI
  Dma_fast_uart_transfer      = 0x10000000, // FUART
  Dma_fast_aic_transfer       = 0x08000000, // FAIC
  Dma_control_trans_halted    = 0x00000008, // HLT
  Dma_control_address_error   = 0x00000004, // AR
  Dma_control_enable          = 0x00000001, // DMAE
};

enum Dma_transfer_count_bits : unsigned
{
  Dma_descriptor_offset_mask  = 0xff000000, // DOA (in DES3)

  Dma_transfer_count_mask     = 0x00ffffff,
  Dma_descriptor_offset_shift = 24,
};

enum Dma_request_source_bits : unsigned
{
  Dma_request_type_mask       = 0x0000003f,
};

enum Dma_control_status_bits : unsigned
{
  Dma_no_descriptor_transfer  = 0x80000000,
  Dma_8word_descriptor        = 0x40000000,
  Dma_copy_offset_mask        = 0x0000ff00,
  Dma_address_error           = 0x00000010,
  Dma_trans_completed         = 0x00000008,
  Dma_trans_halted            = 0x00000004,
  Dma_channel_enable          = 0x00000001,

  Dma_copy_offset_shift       = 8,
};

enum Dma_command_bits : unsigned
{
  Dma_source_address_increment            = 0x800000,
  Dma_source_address_no_increment         = 0x000000,
  Dma_destination_address_increment       = 0x400000,
  Dma_destination_address_no_increment    = 0x000000,

  Dma_source_address_increment_wrap       = 0x200000,
  Dma_destination_address_increment_wrap  = 0x100000,
  Dma_recommended_data_unit_size_mask     = 0x0f0000,
  Dma_source_port_width_mask              = 0x00c000,
  Dma_destination_port_width_mask         = 0x003000,
  Dma_transfer_unit_size_mask             = 0x000f00,

  Dma_trans_unit_size_32_bit              = 0x000000,
  Dma_trans_unit_size_8_bit               = 0x000100,
  Dma_trans_unit_size_16_bit              = 0x000200,
  Dma_trans_unit_size_16_byte             = 0x000300,
  Dma_trans_unit_size_32_byte             = 0x000400,
  Dma_trans_unit_size_64_byte             = 0x000500,
  Dma_trans_unit_size_128_byte            = 0x000600,
  Dma_trans_unit_size_autonomous          = 0x000700,
  Dma_trans_unit_size_external            = 0x000800,

  Dma_source_address_compare_index        = 0x000080,
  Dma_destination_address_compare_index   = 0x000040,
  Dma_stride_enable                       = 0x000004,
  Dma_transfer_irq_enable                 = 0x000002,
  Dma_descriptor_link_enable              = 0x000001,

  Dma_recommended_data_unit_size_shift    = 16,
  Dma_source_port_width_shift             = 14,
  Dma_destination_port_width_shift        = 12,
  Dma_transfer_unit_size_shift            = 8,
};

enum Dma_port_width_values : unsigned
{
  Dma_port_width_32_bit                   = 0,
  Dma_port_width_8_bit                    = 1,
  Dma_port_width_16_bit                   = 2,
};



// Initialise a channel.

Dma_x1600_channel::Dma_x1600_channel(Dma_chip *chip, uint8_t channel,
                                     l4_addr_t start, l4_cap_idx_t irq)
: _chip(chip), _channel(channel), _irq(irq)
{
  _regs = new Hw::Mmio_register_block<32>(start);

  // Initialise the transfer count.

  _regs[Dma_transfer_count] = 0;
}

// Return the closest interval length greater than or equal to the number of
// units given encoded in the request detection interval length field of the
// control/status register.

uint32_t
Dma_x1600_channel::encode_req_detect_int_length(uint8_t units)
{
  static uint8_t lengths[] = {0, 1, 2, 3, 4, 8, 16, 32, 64, 128};
  int i;

  if (!units)
    return 0;

  for (i = 0; i <= 9; i++)
  {
     if (lengths[i] >= units)
       break;
  }

  return i << Dma_recommended_data_unit_size_shift;
}

// Encode the appropriate source port width for the given request type.

uint32_t
Dma_x1600_channel::encode_source_port_width(uint8_t width)
{
  switch (width)
  {
    case 1:
    return Dma_port_width_8_bit << Dma_source_port_width_shift;

    case 2:
    return Dma_port_width_16_bit << Dma_source_port_width_shift;

    default:
    return Dma_port_width_32_bit << Dma_source_port_width_shift;
  }
}

// Encode the appropriate destination port width for the given request type.

uint32_t
Dma_x1600_channel::encode_destination_port_width(uint8_t width)
{
  switch (width)
  {
    case 1:
    return Dma_port_width_8_bit << Dma_destination_port_width_shift;

    case 2:
    return Dma_port_width_16_bit << Dma_destination_port_width_shift;

    default:
    return Dma_port_width_32_bit << Dma_destination_port_width_shift;
  }
}

// Encode the transfer unit size.
// NOTE: This does not handle the external case.

uint32_t
Dma_x1600_channel::encode_transfer_unit_size(uint8_t size)
{
  switch (size)
  {
    case 0:
    return Dma_trans_unit_size_autonomous;

    case 1:
    return Dma_trans_unit_size_8_bit;

    case 2:
    return Dma_trans_unit_size_16_bit;

    case 16:
    return Dma_trans_unit_size_16_byte;

    case 32:
    return Dma_trans_unit_size_32_byte;

    case 64:
    return Dma_trans_unit_size_64_byte;

    case 128:
    return Dma_trans_unit_size_128_byte;

    default:
    return Dma_trans_unit_size_32_bit;
  }
}

// Transfer data between memory locations, returning the number of units that
// should have been transferred.

unsigned int
Dma_x1600_channel::transfer(uint32_t source, uint32_t destination,
                            unsigned int count,
                            bool source_increment, bool destination_increment,
                            uint8_t source_width, uint8_t destination_width,
                            uint8_t transfer_unit_size,
                            int type,
                            l4_addr_t desc_vaddr,
                            l4re_dma_space_dma_addr_t desc_paddr)
{
  // Ensure an absence of address error and halt conditions globally and in this channel.

  if (error() || halted())
  {
    printf("Cleared:%s%s\n", error() ? " error" : "", halted() ? " halted" : "");
    clear_errors();
  }

  // Ensure a zero transfer count for this channel.

  if (_regs[Dma_transfer_count])
    return 0;

  // Disable the channel.

  _regs[Dma_control_status] = _regs[Dma_control_status] & ~Dma_channel_enable;

  // Set transfer count to the number of units.

  unsigned int units = count < Dma_transfer_count_mask ? count : Dma_transfer_count_mask;

  // NOTE: Request detection interval length (for autonomous mode) not considered.

  uint32_t command = (source_increment ? Dma_source_address_increment
                                       : Dma_source_address_no_increment) |
                     (destination_increment ? Dma_destination_address_increment
                                            : Dma_destination_address_no_increment) |
                     encode_source_port_width(source_width) |
                     encode_destination_port_width(destination_width) |
                     encode_transfer_unit_size(transfer_unit_size) |
                     Dma_transfer_irq_enable;

  // Populate the descriptor, largely corresponding to the population of
  // registers when descriptors are not being used.

  if (desc_vaddr)
  {
    // For a descriptor, the descriptor address would be set and the doorbell
    // register field for the channel set.

    // For a descriptor, the actual fields would be populated instead of the
    // command register, descriptor transfer would be indicated in the control/
    // status register along with the appropriate descriptor size indicator.

    struct x1600_dma_descriptor *desc = (struct x1600_dma_descriptor *) desc_vaddr;

    // NOTE: Linking to the same descriptor.

    uint32_t descriptor_offset = 0;

    desc->command = command | Dma_descriptor_link_enable;
    desc->source = source;
    desc->destination = destination;
    desc->transfer_count = (units & Dma_transfer_count_mask) |
                           (descriptor_offset << Dma_descriptor_offset_shift);
    desc->request_source = (enum Dma_x1600_request_type) type;

    // NOTE: Stride not supported yet.

    l4_cache_clean_data((unsigned long) desc_vaddr,
                        (unsigned long) desc_vaddr + sizeof(*desc));

    // Commit the descriptor.

    _regs[Dma_descriptor_address] = desc_paddr;
    _chip->commit_descriptor(_channel);
  }

  // Otherwise, populate the registers for a one-off transfer.

  else
  {
    // Set addresses and transfer count.

    _regs[Dma_source] = source;
    _regs[Dma_destination] = destination;
    _regs[Dma_transfer_count] = units;
    _regs[Dma_request_source] = (enum Dma_x1600_request_type) type;
    _regs[Dma_command] = command;
  }

  // Enable the channel with descriptor transfer configured if appropriate.

  _regs[Dma_control_status] = (desc_vaddr ? Dma_8word_descriptor :
                                            Dma_no_descriptor_transfer) |
                              Dma_channel_enable;

  // Return the number of units to transfer.

  return units;
}

// Wait for a transfer to end, returning the number of units remaining to be
// transferred.

unsigned int
Dma_x1600_channel::wait()
{
  // An interrupt will occur upon completion, the completion flag will be set
  // and the transfer count will be zero.

  unsigned int remaining = 0;

  do
  {
    if (!wait_for_irq(1000000))
      printf("status = %x\n", (uint32_t) _regs[Dma_control_status]);
    else
    {
      remaining = _regs[Dma_transfer_count];
      ack_irq();
      break;
    }
  }
  while (!error() && !halted() && !completed());

  // Reset the channel status.

  _regs[Dma_control_status] = _regs[Dma_control_status] & ~(Dma_channel_enable |
                                  Dma_trans_completed | Dma_address_error |
                                  Dma_trans_halted);
  _regs[Dma_transfer_count] = 0;

  return remaining;
}

// Wait indefinitely for an interrupt request, returning true if one was delivered.

bool
Dma_x1600_channel::wait_for_irq()
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, L4_IPC_NEVER)) && _chip->have_interrupt(_channel);
}

// Wait up to the given timeout (in microseconds) for an interrupt request,
// returning true if one was delivered.

bool
Dma_x1600_channel::wait_for_irq(unsigned int timeout)
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4util_micros2l4to(timeout)))) && _chip->have_interrupt(_channel);
}

// Acknowledge an interrupt condition.

void
Dma_x1600_channel::ack_irq()
{
  _chip->ack_irq(_channel);
}

// Clear error conditions.

void
Dma_x1600_channel::clear_errors()
{
  _regs[Dma_control_status] = _regs[Dma_control_status] & ~(Dma_address_error | Dma_trans_halted);
  _chip->clear_errors();
}

// Return whether a transfer has completed.

bool
Dma_x1600_channel::completed()
{
  return _regs[Dma_control_status] & Dma_trans_completed ? true : false;
}

// Return whether an address error condition has arisen.

bool
Dma_x1600_channel::error()
{
  return _chip->error() || (_regs[Dma_control_status] & Dma_address_error ? true : false);
}

// Return whether a transfer has halted.

bool
Dma_x1600_channel::halted()
{
  return _chip->halted() || (_regs[Dma_control_status] & Dma_trans_halted ? true : false);
}



// Initialise the I2C controller.

Dma_x1600_chip::Dma_x1600_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm)
: _start(start), _end(end), _cpm(cpm)
{
  _regs = new Hw::Mmio_register_block<32>(start);
}

// Enable the peripheral.

void
Dma_x1600_chip::enable()
{
  // Make sure that the DMA clock is available.

  _cpm->start_clock(Clock_dma);

  _regs[Dma_control] = Dma_control_enable;
  while (!(_regs[Dma_control] & Dma_control_enable));
}

// Disable the channel.

void
Dma_x1600_chip::disable()
{
  _regs[Dma_control] = 0;
  while (_regs[Dma_control] & Dma_control_enable);
}

// Obtain a channel object.

Dma_channel *
Dma_x1600_chip::get_channel(uint8_t channel, l4_cap_idx_t irq)
{
  if (channel < 32)
    return new Dma_x1600_channel(this, channel, _start + 0x20 * channel, irq);
  else
    throw -L4_EINVAL;
}

// Return whether an interrupt is pending on the given channel.

bool
Dma_x1600_chip::have_interrupt(uint8_t channel)
{
  return _regs[Dma_irq_pending] & (1UL << channel) ? true : false;
}

// Acknowledge an interrupt condition on the given channel.

void
Dma_x1600_chip::ack_irq(uint8_t channel)
{
  _regs[Dma_irq_pending] = _regs[Dma_irq_pending] & ~(1UL << channel);
}

// Clear error conditions.

void
Dma_x1600_chip::clear_errors()
{
  _regs[Dma_control] = _regs[Dma_control] & ~(Dma_control_address_error | Dma_control_trans_halted);
}

// Return whether an address error condition has arisen.

bool
Dma_x1600_chip::error()
{
  return _regs[Dma_control] & Dma_control_address_error ? true : false;
}

// Return whether a transfer has halted.

bool
Dma_x1600_chip::halted()
{
  return _regs[Dma_control] & Dma_control_trans_halted ? true : false;
}

void
Dma_x1600_chip::commit_descriptor(uint8_t channel)
{
  _regs[Dma_doorbell_set] = (1 << channel);
}

Dma_chip *x1600_dma_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm)
{
  return new Dma_x1600_chip(start, end, cpm);
}



// C language interface functions.

void *x1600_dma_init(l4_addr_t start, l4_addr_t end, void *cpm)
{
  return (void *) x1600_dma_chip(start, end, static_cast<Cpm_chip *>(cpm));
}

void x1600_dma_disable(void *dma_chip)
{
  static_cast<Dma_chip *>(dma_chip)->disable();
}

void x1600_dma_enable(void *dma_chip)
{
  static_cast<Dma_chip *>(dma_chip)->enable();
}

void *x1600_dma_get_channel(void *dma, uint8_t channel, l4_cap_idx_t irq)
{
  return static_cast<Dma_chip *>(dma)->get_channel(channel, irq);
}

unsigned int x1600_dma_transfer(void *dma_channel,
                                uint32_t source, uint32_t destination,
                                unsigned int count,
                                int source_increment, int destination_increment,
                                uint8_t source_width, uint8_t destination_width,
                                uint8_t transfer_unit_size,
                                enum Dma_x1600_request_type type)
{
  return static_cast<Dma_x1600_channel *>(dma_channel)->transfer(source,
    destination, count, source_increment, destination_increment, source_width,
    destination_width, transfer_unit_size, type);
}

unsigned int x1600_dma_wait(void *dma_channel)
{
  return static_cast<Dma_x1600_channel *>(dma_channel)->wait();
}
